import pytest
from django.utils import timezone
from model_bakery import baker
from sage_mailbox.models import EmailMessage


@pytest.mark.django_db
class TestEmailMessageModel:

    def test_sanitize_message_id(self):
        """Test the sanitize_message_id class method."""
        msg_id = "<test@example.com>"
        sanitized_id = EmailMessage.sanitize_message_id(msg_id)
        assert sanitized_id == "<test@example.com>"

        invalid_msg_id = "invalid_id"
        sanitized_invalid_id = EmailMessage.sanitize_message_id(invalid_msg_id)
        assert sanitized_invalid_id is None

    def test_to_dataclass(self):
        """Test conversion from EmailMessage to dataclass."""
        email = baker.make(EmailMessage, subject="Test subject")
        email_dc = email.to_dataclass()

        assert email_dc.subject == "Test subject"

    def test_has_attachments_method(self):
        """Test has_attachments method for EmailMessage."""
        email = baker.make(EmailMessage)
        baker.make('sage_mailbox.Attachment', email_message=email, _quantity=2)

        assert email.has_attachments() is True

    def test_get_summary(self):
        """Test get_summary method."""
        email = baker.make(
            EmailMessage,
            subject="Test",
            from_address="from@example.com",
            date=timezone.now()
        )
        summary = email.get_summary()

        assert summary["subject"] == "Test"
        assert summary["from"] == "from@example.com"
        assert "has_attachments" in summary



import imaplib
import logging

from django.conf import settings
from django.core.checks import Error, Warning, register

from sage_mailbox.conf import imap_settings
from sage_mailbox.exc import (
    IMAPAuthenticationError,
    IMAPConfigurationError,
    IMAPConnectionError,
    IMAPUnexpectedError,
)

logger = logging.getLogger(__name__)


@register()
def check_imap_config(app_configs, **kwargs):
    """
    Check the IMAP configuration and other required settings for the application.

    This function verifies that all required IMAP settings, Django apps, and certain
    settings are present. It also attempts to establish a connection to the IMAP server
    to ensure the settings are correct. Any errors encountered during these checks
    are returned.

    Parameters
    ----------
    app_configs : dict
        The application configurations.
    **kwargs
        Additional keyword arguments.

    Returns
    -------
    list of Error or Warning
        A list of Error or Warning objects representing any configuration or connection
        errors found.

    Raises
    ------
    IMAPConfigurationError
        If any required IMAP configuration settings are missing.
    IMAPAuthenticationError
        If the authentication with the IMAP server fails.
    IMAPConnectionError
        If there is a connection error with the IMAP server.
    IMAPUnexpectedError
        If an unexpected error occurs during the IMAP configuration check.

    Examples
    --------
    >>> errors = check_imap_config(app_configs)
    >>> if errors:
    ...     for error in errors:
    ...         print(error)
    """
    errors = []

    # Check that required apps are installed
    required_apps = ["django.contrib.sites", "django_jsonform"]
    for app in required_apps:
        if app not in settings.INSTALLED_APPS:
            errors.append(
                Error(
                    f"The required app '{app}' is not installed.",
                    id="sage_integration.E001",
                )
            )

    # Check that SITE_ID is set
    if not hasattr(settings, "SITE_ID"):
        errors.append(
            Error(
                "SITE_ID is not set in settings.",
                id="sage_integration.E002",
            )
        )
    elif not isinstance(settings.SITE_ID, int):
        errors.append(
            Error(
                "SITE_ID must be an integer.",
                id="sage_integration.E003",
            )
        )

    # Check that upload-related settings are set
    upload_settings = ["MEDIA_URL", "MEDIA_ROOT", "FILE_UPLOAD_HANDLERS"]
    for setting in upload_settings:
        if not hasattr(settings, setting):
            errors.append(
                Error(
                    f"The required setting '{setting}' is not set.",
                    id="sage_integration.E008",
                )
            )

    # Function to get IMAP settings
    def get_imap_settings():
        return {
            "IMAP_SERVER_DOMAIN": imap_settings.IMAP_SERVER_DOMAIN,
            "IMAP_SERVER_PORT": imap_settings.IMAP_SERVER_PORT,
            "IMAP_SERVER_USER": imap_settings.IMAP_SERVER_USER,
            "IMAP_SERVER_PASSWORD": imap_settings.IMAP_SERVER_PASSWORD,
        }

    # Function to check missing configurations
    def check_missing_configs(settings):
        missing = [key for key, value in settings.items() if not value]
        if missing:
            raise IMAPConfigurationError(
                f"IMAP configuration settings are missing: {', '.join(missing)}."
            )

    # Check IMAP configuration settings
    try:
        imap_settings_dict = get_imap_settings()
        check_missing_configs(imap_settings_dict)
    except IMAPConfigurationError as e:
        errors.append(
            Error(
                str(e),
                id="sage_integration.E004",
            )
        )

    # Check IMAP connection
    try:
        mail = imaplib.IMAP4_SSL(
            imap_settings_dict["IMAP_SERVER_DOMAIN"],
            imap_settings_dict["IMAP_SERVER_PORT"],
        )
        mail.login(
            imap_settings_dict["IMAP_SERVER_USER"],
            imap_settings_dict["IMAP_SERVER_PASSWORD"],
        )
        mail.logout()
    except imaplib.IMAP4.error as e:
        if "authentication failed" in str(e).lower():
            errors.append(
                Error(
                    str(IMAPAuthenticationError()),
                    id="sage_integration.E007",
                )
            )
        else:
            errors.append(
                Error(
                    str(IMAPConnectionError(detail=str(e))),
                    id="sage_integration.E005",
                )
            )
    except Exception as e:
        errors.append(
            Error(
                str(IMAPUnexpectedError(detail=str(e))),
                id="sage_integration.E006",
            )
        )

    return errors



import pytest
from django.utils import timezone
from model_bakery import baker
from datetime import datetime
from sage_mailbox.models import EmailMessage, Flag, Mailbox
from autoslug import AutoSlugField
from django.db.models import F


# Register the custom generator for AutoSlugField
def slug_generator():
    return 'test-slug'

baker.generators.add(AutoSlugField, slug_generator)


@pytest.mark.django_db
class TestEmailMessageQuerySet:
    
    def test_total_attachments(self):
        """Test total attachment count for each email."""
        email = baker.make(EmailMessage, _quantity=3)
        attachment = baker.make('sage_mailbox.Attachment', email_message=email[0], _quantity=5)
        attachment2 = baker.make('sage_mailbox.Attachment', email_message=email[1], _quantity=2)

        queryset = EmailMessage.objects.total_attachments()

        assert queryset.get(id=email[0].id).total_attachments == 5
        assert queryset.get(id=email[1].id).total_attachments == 2
        assert queryset.get(id=email[2].id).total_attachments == 0

    def test_has_attachments(self):
        """Test annotation to check if emails have attachments."""
        email = baker.make(EmailMessage, _quantity=3)
        baker.make('sage_mailbox.Attachment', email_message=email[0])

        queryset = EmailMessage.objects.has_attachments()
        assert queryset.get(id=email[0].id).has_attachments is True
        assert queryset.get(id=email[1].id).has_attachments is False

    def test_select_related_mailbox(self, django_assert_num_queries):
        """Test select_related optimization for mailbox."""
        mailbox = baker.make(Mailbox)
        email = baker.make(EmailMessage, mailbox=mailbox)

        # Ensure that select_related is optimizing the query by fetching related objects in one query
        with django_assert_num_queries(1):  # One query should fetch both EmailMessage and Mailbox
            result_email = EmailMessage.objects.select_related_mailbox().get(id=email.id)
            assert result_email.mailbox == mailbox  # Access the mailbox without triggering another query

    def test_list_attachments(self):
        """Test prefetching attachments."""
        email = baker.make(EmailMessage)
        baker.make('sage_mailbox.Attachment', email_message=email, _quantity=3)
        queryset = EmailMessage.objects.list_attachments()

        assert queryset.count() == 1
        assert queryset[0].attachments.count() == 3

    def test_unread(self):
        """Test filtering unread emails."""
        baker.make(EmailMessage, is_read=False, _quantity=2)
        baker.make(EmailMessage, is_read=True, _quantity=2)

        assert EmailMessage.objects.unread().count() == 2

    def test_flagged(self):
        """Test filtering flagged emails."""
        baker.make(EmailMessage, is_flagged=True, _quantity=3)
        baker.make(EmailMessage, is_flagged=False, _quantity=2)

        assert EmailMessage.objects.flagged().count() == 3


@pytest.mark.django_db
class TestEmailMessageManager:
    
    def test_total_attachments_manager(self):
        """Test total attachments through manager."""
        email = baker.make(EmailMessage)
        baker.make('sage_mailbox.Attachment', email_message=email, _quantity=2)

        queryset = EmailMessage.objects.total_attachments()
        assert queryset.get(id=email.id).total_attachments == 2

    def test_has_attachments_manager(self):
        """Test has_attachments through manager."""
        email = baker.make(EmailMessage)
        baker.make('sage_mailbox.Attachment', email_message=email)

        queryset = EmailMessage.objects.has_attachments()
        assert queryset.get(id=email.id).has_attachments is True

    def test_unread_manager(self):
        """Test unread emails through manager."""
        baker.make(EmailMessage, is_read=False)
        assert EmailMessage.objects.unread().count() == 1

    def test_flagged_manager(self):
        """Test flagged emails through manager."""
        baker.make(EmailMessage, is_flagged=True)
        assert EmailMessage.objects.flagged().count() == 1



import pytest

from django.core.exceptions import ValidationError

from sage_mailbox.validators import FolderNameValidator


class TestFolderNameValidator:

    def setup_method(self):
        self.validator = FolderNameValidator()

    @pytest.mark.parametrize("folder_name", [
        "MyFolder",           # letters
        "Folder123",          # letters and numbers
        "Folder-Name_123",    # hyphen and underscore
        "Folder.Name",        # dot in name
        "A" * 255             # exactly 255 characters
    ])
    def test_valid_folder_names(self, folder_name):
        # These should pass validation
        result = self.validator(folder_name)
        assert result is None

    @pytest.mark.parametrize("folder_name", [
        "",                          # Empty string
        "A" * 256,                   # Exceeds 255 characters
    ])
    def test_invalid_folder_names_length(self, folder_name):
        with pytest.raises(ValidationError) as exc_info:
            self.validator(folder_name)
        assert exc_info.value.messages[0] == "Folder name must be between 1 and 255 characters long."

    @pytest.mark.parametrize("folder_name", [
        "Invalid@Folder",            # Invalid special characters
        "Folder#Name",               # Invalid special character
        "Folder Name With Spaces",   # Spaces in name
        "-Invalid",                  # Starts with invalid character
        "Invalid-"                   # Ends with invalid character
    ])
    def test_invalid_folder_names_characters(self, folder_name):
        with pytest.raises(ValidationError) as exc_info:
            self.validator(folder_name)
        assert exc_info.value.messages[0] == "Folder name contains invalid characters. Allowed characters are letters, numbers, underscore, hyphen, and dot. Spaces are not allowed."

    @pytest.mark.parametrize("folder_name", [
        "A",                         # Exactly 1 character
        "A" * 255                    # Exactly 255 characters
    ])
    def test_edge_cases(self, folder_name):
        # Edge case for exactly 1 and 255 characters, should pass validation
        result = self.validator(folder_name)
        assert result is None



import pytest

from django.core.exceptions import ValidationError

from sage_mailbox.validators import CommaSeparatedEmailValidator


class TestCommaSeparatedEmailValidator:

    def setup_method(self):
        self.validator = CommaSeparatedEmailValidator()

    @pytest.mark.parametrize("emails", [
        "test@example.com",                          # Single valid email
        "test1@example.com, test2@example.com",      # Multiple valid emails
        "test1@example.com,   test2@example.com",    # Multiple valid emails with spaces
    ])
    def test_valid_emails(self, emails):
        # These should pass validation
        result = self.validator(emails)
        assert result is None

    @pytest.mark.parametrize("emails", [
        "invalid-email",                             # Single invalid email
        "valid@example.com, invalid-email",          # One valid, one invalid
        "invalid-email1, invalid-email2",            # Both invalid
        "test@",                                     # Missing domain
        "@example.com",                              # Missing local part
    ])
    def test_invalid_emails(self, emails):
        with pytest.raises(ValidationError):
            self.validator(emails)

    @pytest.mark.parametrize("emails", [
        "",                                          # Empty string
        "   test@example.com   ",                    # Valid email with leading/trailing spaces
        "valid@example.com,   ",                     # Valid email with trailing comma and space
    ])
    def test_edge_cases(self, emails):
        # These should pass validation
        result = self.validator(emails)
        assert result is None



# import pytest
# from unittest.mock import patch, MagicMock
# from dateutil import parser
# from sage_mailbox.repository.service import EmailSyncService


# from sage_mailbox.models import EmailMessage as DjangoEmailMessage
# from sage_mailbox.models import Mailbox as DjangoMailbox
# from sage_mailbox.models import Attachment as DjangoAttachment
# from sage_mailbox.models import Flag as DjangoFlag


# from sage_imap.services import IMAPClient, IMAPMailboxService
# from sage_imap.helpers.enums import Flag, MessagePart
# from sage_imap.helpers.search import IMAPSearchCriteria
# from sage_imap.models.message import MessageSet
# from sage_imap.models.email import EmailMessage

# # Sample data for tests
# SAMPLE_EMAIL = EmailMessage(
#     uid="123",
#     message_id="msg-123",
#     subject="Test Subject",
#     from_address="from@example.com",
#     to_address=["to@example.com"],
#     cc_address=[],
#     bcc_address=[],
#     date="01-Jan-2021",
#     raw=b"Raw message content",  # Ensure raw is bytes
#     plain_body="Plain body content",
#     html_body="<html>HTML body content</html>",
#     size=1024,
#     flags=[Flag.SEEN, Flag.FLAGGED],
#     headers={"Message-ID": "<msg-123>", "subject": "Test Subject"},
#     attachments=[]
# )


# @pytest.fixture
# def email_sync_service():
#     """Fixture to instantiate the EmailSyncService."""
#     return EmailSyncService(host="imap.example.com", username="user", password="password")


# @pytest.fixture
# def mocked_mailbox():
#     """Fixture to mock IMAPMailboxService."""
#     with patch("sage_imap.services.IMAPMailboxService") as mock_mailbox:
#         yield mock_mailbox


# @pytest.fixture
# def mocked_client():
#     """Fixture to mock IMAPClient."""
#     with patch("sage_imap.services.IMAPClient") as mock_client:
#         yield mock_client


# @pytest.fixture
# def mocked_django_models():
#     """Fixture to mock Django model interactions."""
#     with patch("sage_mailbox.models.EmailMessage.objects.update_or_create") as mock_update_or_create, \
#          patch("sage_mailbox.models.Mailbox.objects.get_or_create") as mock_get_or_create, \
#          patch("sage_mailbox.models.Attachment.objects.update_or_create") as mock_attachment_create, \
#          patch("sage_mailbox.models.Flag.objects.get_or_create") as mock_flag_create:
#         mock_update_or_create.return_value = (MagicMock(), True)  # Mocking the return values
#         mock_get_or_create.return_value = (MagicMock(), True)  # Mock Mailbox get_or_create
#         mock_attachment_create.return_value = (MagicMock(), True)  # Mock Attachment creation
#         mock_flag_create.return_value = (MagicMock(), True)  # Mock Flag creation
#         yield mock_update_or_create, mock_get_or_create, mock_attachment_create, mock_flag_create


# class TestEmailSyncService:
#     def test_service_initialization(self, email_sync_service):
#         """Test the initialization of EmailSyncService."""
#         assert email_sync_service.host == "imap.example.com"
#         assert email_sync_service.username == "user"
#         assert email_sync_service.password == "password"

#     def test_fetch_emails_from_default_inbox(self, email_sync_service, mocked_client, mocked_mailbox, mocked_django_models):
#         """Test fetching emails from the default folder 'INBOX'."""
#         mock_mailbox_instance = mocked_mailbox.return_value
#         mock_mailbox_instance.search.return_value = [1, 2, 3]  # Fake message IDs
#         mock_mailbox_instance.fetch.return_value = [SAMPLE_EMAIL]

#         # Mock Django mailbox retrieval/creation
#         mock_django_email_message, mock_mailbox_obj = mocked_django_models[1], mocked_django_models[0]
#         mock_django_email_message.filter.return_value.order_by.return_value.first.return_value = None

#         result = email_sync_service.fetch_and_save_emails()

#         assert result == {"created_emails": 1, "created_attachments": 0}
#         mock_mailbox_instance.search.assert_called_once_with(IMAPSearchCriteria.ALL)
#         mock_mailbox_instance.fetch.assert_called_once()

#     def test_fetch_emails_with_latest_date(self, email_sync_service, mocked_client, mocked_mailbox, mocked_django_models):
#         """Test fetching emails when there's already synced emails in the database."""
#         mock_mailbox_instance = mocked_mailbox.return_value
#         mock_mailbox_instance.search.return_value = [1, 2, 3]  # Fake message IDs
#         mock_mailbox_instance.fetch.return_value = [SAMPLE_EMAIL]

#         # Mock latest email creation date
#         latest_email = MagicMock()
#         latest_email.created_at = parser.parse("2021-01-01")
#         mock_django_email_message, _ = mocked_django_models[1], mocked_django_models[0]
#         mock_django_email_message.filter.return_value.order_by.return_value.first.return_value = latest_email

#         result = email_sync_service.fetch_and_save_emails()

#         assert result == {"created_emails": 1, "created_attachments": 0}
#         mock_mailbox_instance.search.assert_called_once()
#         mock_mailbox_instance.fetch.assert_called_once()

#     def test_handle_invalid_email_date(self, email_sync_service, mocked_django_models):
#         """Test saving email when email.date is invalid."""
#         invalid_email = SAMPLE_EMAIL
#         invalid_email.date = "invalid-date"

#         mock_django_email_message, _ = mocked_django_models[1], mocked_django_models[0]
#         email_sync_service.create_or_update_email(invalid_email, MagicMock())

#         mock_django_email_message.update_or_create.assert_called_once()

#     def test_save_email_with_attachments(self, email_sync_service, mocked_django_models):
#         """Test saving email with attachments."""
#         attachment = MagicMock(filename="file.txt", payload=b"file_content", content_type="text/plain")
#         email_with_attachments = SAMPLE_EMAIL
#         email_with_attachments.attachments = [attachment]

#         mock_django_email_message, _ = mocked_django_models[1], mocked_django_models[0]
#         result = email_sync_service.create_or_update_email(email_with_attachments, MagicMock())

#         mock_django_email_message.update_or_create.assert_called_once()
#         assert result == (True, 1)  # One email and one attachment created

#     def test_error_handling_in_fetch_and_save_emails(self, email_sync_service, mocked_mailbox):
#         """Test error handling during email fetching."""
#         mock_mailbox_instance = mocked_mailbox.return_value
#         mock_mailbox_instance.search.side_effect = Exception("IMAP error")

#         result = email_sync_service.fetch_and_save_emails()
#         assert result == {"created_emails": 0, "created_attachments": 0}

#     def test_handle_flags_for_email(self, email_sync_service, mocked_django_models):
#         """Test handling flags for an email."""
#         mock_django_flag, _ = mocked_django_models[3], mocked_django_models[0]

#         email_sync_service.handle_flags(MagicMock(), [Flag.SEEN, Flag.FLAGGED])

#         assert mock_django_flag.call_count == 2  # Two flags should be created or retrieved

#     def test_handle_no_attachments(self, email_sync_service, mocked_django_models):
#         """Test saving an email with no attachments."""
#         mock_django_email_message, _ = mocked_django_models[1], mocked_django_models[0]

#         result = email_sync_service.create_or_update_email(SAMPLE_EMAIL, MagicMock())

#         mock_django_email_message.update_or_create.assert_called_once()
#         assert result == (True, 0)  # One email and zero attachments created



from autoslug import AutoSlugField
from django.db import models
from django.utils.translation import gettext_lazy as _

from sage_mailbox.models.mixins import TimestampMixin
from sage_mailbox.utils import map_to_standard_name
from sage_mailbox.validators import validate_folder_name


def custom_slugify(value):
    return value.replace(".", "-").replace(" ", "-")


class StandardMailboxNames(models.TextChoices):
    INBOX = "INBOX", _("Inbox")
    SENT = "SENT", _("Sent")
    DRAFTS = "DRAFTS", _("Drafts")
    SPAM = "SPAM", _("Spam")
    JUNK = "JUNK", _("Junk")
    TRASH = "TRASH", _("Trash")
    CUSTOM = "CUSTOM", _("Custom")


IMAP_TO_STANDARD_MAP = {
    "INBOX": StandardMailboxNames.INBOX,
    "Sent Items": StandardMailboxNames.SENT,
    "Sent": StandardMailboxNames.SENT,
    "Drafts": StandardMailboxNames.DRAFTS,
    "Junk": StandardMailboxNames.JUNK,
    "Spam": StandardMailboxNames.SPAM,
    "Trash": StandardMailboxNames.TRASH,
    "Deleted Items": StandardMailboxNames.TRASH,
}


class Mailbox(TimestampMixin):
    name = models.CharField(
        max_length=255,
        unique=True,
        validators=[validate_folder_name],
        verbose_name=_("Name"),
        help_text=_("The unique name of the mailbox."),
        db_comment="The unique name identifying the mailbox.",
    )
    folder_type = models.CharField(
        max_length=255,
        verbose_name=_("Folder Type"),
        null=True,
        blank=False,
        help_text=_("The standardized type of the mailbox."),
        db_comment="The standardized type of the mailbox.",
        choices=StandardMailboxNames.choices,
        default=StandardMailboxNames.CUSTOM,
    )
    slug = AutoSlugField(
        verbose_name=_("Slug"),
        max_length=255,
        unique=True,
        always_update=True,
        allow_unicode=True,
        slugify=custom_slugify,
        populate_from="name",
        help_text=_("The unique slug for the mailbox, generated from the name."),
        db_comment="The unique slug generated from the mailbox name.",
    )

    class Meta:
        verbose_name = _("Mailbox")
        verbose_name_plural = _("Mailboxes")
        indexes = [
            models.Index(fields=["name"], name="idx_mailbox_name"),
            models.Index(fields=["slug"], name="idx_mailbox_slug"),
        ]
        db_table = "sage_mailbox"
        db_table_comment = (
            "Model representing a mailbox which can be nested under another mailbox."
        )

    def save(self, *args, **kwargs):
        # Map the name to the standard folder type
        if not self.pk:
            self.folder_type = map_to_standard_name(self.name)

        super().save(*args, **kwargs)

    def __str__(self):
        return str(self.name)
